/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>.
 * Copyright (c) 2001, 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 850 $ $LastChangedDate: 2002-09-01 22:13:49 +0200 (Sun, 01 Sep 2002) $
 *
 * Ast - Gont Abstract Syntax Trees.
 *
 * Definition of datatypes describing Gont programs throughout the compiler. 
 * This module exports no functions.
 */
 
/* Identifier.  */
typedef string id;

/* Location in source file.  */
struct location {
	string file;
	int first_line;
	int first_column;
	int last_line;
	int last_column;
}

/* Several constructs need location information.  This type is used to 
   provide it.  Naming convention is nl_foo as 'a, and 
   typedef <nl_foo>located foo; later.  nl_ stands for non_located_.  */
struct <'a>located {
	'a a;
	location loc;
}

/* Names of different kinds (names of variables, types, struct fields, 
   labels etc.  Basically everything that is expressed by ID in language
   grammar.) are assigned objects of type <'a>symbol.  In <'x>bindable, 
   'x is always <'a>symbol, for some 'a.  <'a>bindable types is stored in 
   several records under *_symbol field.  (Simple|Qualified)Name(...) gets
   replaced by Symbol(...) after name binding for most of *_symbol fields.
   If it's not the case it's marked in the field description.  */
union <'a>bindable {
	string Simple_name;
	*(string, string) Qualified_name;
	'a Symbol;
}

struct nt_record {
	<ty>list parms;	// Actual types to be substituted to
			// tydef.Symbol.tyvars
	<ty_symbol>bindable tydef;
}

/* tyvars is generally filled by the parser using simple_fv function
 * called on parameters and return type. However this is wrong for nested
 * functions (for example `f('a x) { g('a x, 'b y) {} }', tyvars for g
 * should be ('b), but it is ('a, 'b)), this is fixed by Gonttypes module. 
 * That's why we need tyvars to be mutable. */
 
struct ft_record {
	ty ret_type;
	<*(ty, <id>option)>list parms;
	<id>list tyvars;
}

struct <'a>field {
	<'a>bindable fld_symbol;
	ty fld_type;
	location fld_loc;
}

struct xtyvar {
	// without ', i.e. for 'ble it's "ble"
	id name;
	// used for unification
	<ty>option t;
	// used in checking for cycles etc, in normal circumstances should
	// be false
	bool flag;
}

// Types used by Gont compiler.
union ty {
	void Int;		// integer of the same size as pointer
	void Void;		// unit type
	void Bool;		// boolean, same size as Int
	void Float;		// C's double
	void String;		// Pointer to struct string$
	<ty>list Tuple;	// void **
	ft_record Func;
	id Tyvar;
	xtyvar X_tyvar;		// used for unification and weak type variables
	<<sf_symbol>field>list Struct;
	<<sf_symbol>field>list Opt_struct;
	<<var_symbol>field>list Union;
	void Abstract_type;	// Used when outside module defines `type t'
	nt_record Named_type;	// <int>list, some_struct
	void Exn;		// Type of exception value.
}

struct <'a>symbol {
	id module;
	id name;
	ty stype;
	location sloc;
	'a a;
}

// *** var_symbol:
// This includes global and local variables and functions. Also union
// fields and exceptions go here.

// There can be only void fields here. If you're changing it, look at
// Module::{,un}xmlize_props too
union var_symbol_property {
	void Is_in_closure;
	void Is_union_field;
	void Is_lvalue;
	void Is_initialized;
	void Is_global_var;
	void Is_exception;
	void Is_exported;
	void Is_extern_c;

	// generated only by codegen, or by other modules, but
	// just for local symbols (i.e. Module doesn't need to worry
	// about them)
	void Is_prototyped;
	void Is_in_pattern;
	void Is_closurized;
	void Is_guard_local;
	void Is_let_variable;
	void Is_statically_allocated;
	void Is_globally_safe;
	void Is_used;
	void Is_formal_parm;
}

struct var_symbol_data {
	// Level of function nesting. 0 for globals, -1 for symbols from other
	// modules, see gontbind.ml for more info.
	int fnc_level;
	
	// Same as above, but s/function/scope/.
	int scope_level;

	// How to call it from Ksi level
	id extern_name;
	
	// special properties (flags) symbol might have.
	int properties;
	<ty_symbol>option parent_union;
}

typedef <var_symbol_data>symbol var_symbol;

// *** sf_symbol -- struct field symbol
struct sf_symbol_data {
	// typedef, this field was used in.
	ty_symbol sf_tydef;
}

typedef <sf_symbol_data>symbol sf_symbol;

// *** li_symbol
// Labels assigned to loops, that can be used by break and continue stmts.
struct li_symbol_data {
	int loop_level;
	bool is_used;
}

typedef <li_symbol_data>symbol li_symbol;

// *** ty_symbol
// used for typedefs
// Actual defined type (Struct or Union in most cases) is in stype.
// module is "" for current module, until binder.
struct ty_symbol_data {
	// Parameters of defined type.
	<id>list tysym_tyvars;
}

typedef <ty_symbol_data>symbol ty_symbol;


/*** Expressions. ***/

// Type of binary operator.
union op2 { 
	void Plus;
	void Minus;
	void Mult;
	void Div;
	void Mod;
	void Eq;
	void Ne;
	void Lt;
	void Gt;
	void Ge;
	void Le;
	void And_if;
	void Or_if;

	void Bit_and;
	void Bit_or;
	void Bit_xor;
	void Bit_lshift;
	void Bit_rshift;
}
  
// Type of unary operators.
union op1 {
	void Negate;
	void Not;
	void Pre_dec;
	void Pre_inc;
	void Post_dec;
	void Post_inc;
	void Bit_not;
}

/* Formal parameter of function. parm_symbol set by Gontbind. Note that it
 * still can be in closure (it is assigned there at the function startup). */
struct vardef {
	<var_symbol>bindable var_symbol;
	ty var_type;
}

struct namedexpr {
	<var_symbol>bindable nx_symbol;
	expr nx_expr;
}

struct sini_expr {
	<sf_symbol>bindable sx_symbol;
	expr sx_expr;
}

struct ref_record {
	<var_symbol>bindable ref_symbol;
	/* How deep in closure chain should we go to get to
	 * the value. If level == 0, the value is defined
	 * at the same function level, and we don't need to
	 * go anywhere to get to it :^), however if 
	 * ref_symbol#in_closure, we need to access it through
	 * cl$ pointer. */
	int ref_level;
}

struct sf_binding {
	<ty_symbol>list sfb_tydefs;
}

// Expression and statement representation trees.
union raw_expr {
	*(expr, expr) Assign;
	*(op2, expr, expr) Simple_binop;
	*(op1, expr) Simple_unop;
	ref_record Ref; 		// reference to variable
	*(expr, <expr>list) Call; 	// functional variable * parameters
	bool Bool_const;
	Int64::t Int_const;
	float Float_const;
	void Void_const;			// ()
	string String_const;
	void Null_const;			// null
	<expr>list Tuple_expr;
	namedexpr Union_expr;
	lambda_record Lambda;
	*(<sini_expr>list, sf_binding) Struct_expr;	// {foo=1, bar=x+y}
	*(expr, <sini_expr>list, sf_binding) Struct_copy;	
					// foo with {bar=1, qux=3}
	*(sini_expr, sf_binding) Dot_expr;		
					// foo.bar, symbol set typechecker,
					// not binder!
	// we represent "x += 3" as "x = Assign_target + 3"
	void Assign_target;
}

struct nl_expr {
	raw_expr raw;
	// Filled by Gonttypes, but don't use it directly, call
        // Gonttypes.type_of instead.
	<ty>option expr_type;
}

typedef <nl_expr>located expr;

/*** Patterns. ***/

struct named_pattern {
	<var_symbol>bindable np_symbol;
	pattern np_pattern;
}

struct sp_entry {
	<sf_symbol>bindable sp_symbol;
	pattern sp_pattern;
}

struct sp_record {
	<sp_entry>list sp_entries;
	sf_binding sp_bindings;
	ty sp_type;
}

union nl_pattern {
	void Void_pattern;
	vardef Var_pattern;
	*(pattern, vardef) As_pattern;
	expr Expr_pattern;
	<pattern>list Tuple_pattern;
	named_pattern Union_pattern;
	sp_record Struct_pattern;
}

typedef <nl_pattern>located pattern;

/*** Statements. ***/

union break_kind {
	void Break_kind;
	void Continue_kind;
}

struct break_record {
	<li_symbol>bindable break_symbol;
	break_kind break_kind;
}

struct loop_record {
	// If unlabeled, it's "", Symbol is of kind Loop_id_kind.
	<li_symbol>bindable ll_symbol;
	stmt ll_body;
	// s3 in for(s1;s2;s3), executed, even is continue is called.
	stmt ll_inc;	
}

struct lambda_record {
	ty lambda_type;
	stmt lambda_body;
	// all filled by name binder
	<var_symbol>list lambda_closure_vars;
	<var_symbol>list lambda_parms;
	bool lambda_needs_closure;
}

struct case_clause {
	// expr is condition, it's Bool_const(true) if there was none
	<*(pattern, expr)>list cases;
	stmt body;

	// filled by name binder, this doesn't include pattern variables
	// with Is_guard_local set.
	<var_symbol>list patvars;
}

struct try_record {
	stmt try_body;
	<case_clause>list try_with;
	stmt try_finally;
}

union nl_stmt {
	void Skip;
	expr Expr;              /* Expression statement -- evaluate expression
                                 * and forget the value. */
				 
	expr Return;             /* Return given value from function. */
	
	vardef Vardef;      	/* int foo; */
	
  	<stmt>list Block;    /* { ... } */
	
  	*(stmt, stmt) Pair;     /* Doesn't introduce new scope, simply
                                 * groups statements, it is introduced
                                 * by few parser rules, however there is
                                 * no user-mode way to get it. */
				 
	  /* Stores base expression which is matched to first fitting 
	   * pattern in (pattern*stmt) list and its stmt is executed. */
	*(expr, <case_clause>list) Switch;

	*(<*(pattern, expr)>list, stmt) Let;
	<*(pattern, expr)>list Flying_let;
	
	*(expr, stmt, stmt) If;     /* if (expr) stmt else stmt */
	
  /* All loops are constructed from these two terms. for(a;b;c)X loop is
   * split into Block(a,Loop(Pair(Pair(Break_if_not(b),X),c))), while(c)X
   * into Loop(Pair(Break_if_not(c),X)), and so on. */
	loop_record Loop;
	expr Break_if_not;
	
  /* This if for regular, user-supplied, breaks and continues. */
	break_record Break;
	
  /* Local open -- import all symbols from given module into current 
   * namespace, however note, symbols local to module always override
   * external symbols! */
	id Open;
	
	try_record Try;
	
	expr Raise;                     /* raise Foo("bar", baz) */
}

typedef <nl_stmt>located stmt;

/*** Declarations ***/

struct fn_record {
	<var_symbol>bindable fn_symbol;
        string fn_export_as;

	lambda_record fn_lambda;
}

union nl_decl {
	fn_record Func_decl;
	vardef Global_decl;
	*(string, vardef) Extern_c;
	vardef Func_proto;
  	id Top_level_open;
	ty_symbol Tydef;
	vardef Exception_decl;
	*(id, lambda_record) Section_decl;
}

typedef <nl_decl>located decl;
