/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * Copyright (c) 2001, 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 730 $ $LastChangedDate: 2002-08-31 09:23:01 +0200 (Sat, 31 Aug 2002) $
 *
 * Ksi - Interface to the Ksi intermediate language.
 */

open Ksi;

typedef string id;

id make_id(string s)
{
    return s;
}

ty ptroff()
{
    return Macro_type("ptr_off");
}

ty uptroff()
{
    return Macro_type("uptr_off");
}

void output_header(Stdio::FILE out)
{
    Stdio::fputs(out, "// generated by *Gont* interface ;)\n"
		      "!version 1\n"
		      "\n"
		      "[\n");
}


void output_trailer(Stdio::FILE out)
{
    Stdio::fputs(out, "]\n\n// EOF\n");
}

int infty;
int line_len;
bool fast_writer;
<tstring>Array::t indents;

union tstring {
    string Leaf;
    *(tstring, tstring, int) Pair;  // int is summaric length
}

<tstring>list convert_string(string s)
{
    int max_len = 50;
    string specials = "\r\n\t\"\\", subst = "rnt\"\\";
    char backslash = '\\';
    
    int pos = 0;
    string tmp = String::create(max_len + 3);
    <tstring>list ret = Nil;

    void append(char c)
    {
	if (String::contains(specials, c)) {
	    int idx = String::index(specials, c);
	    char sbst = String::get(subst, idx);
	    String::set(tmp, pos++, backslash); 
	    String::set(tmp, pos++, sbst);
	} else {
	    String::set(tmp, pos++, c); 
	}

	if (pos >= max_len) {
	    string ss = "\"" + String::sub(tmp, 0, pos) + "\".";
	    ret = Cons(Leaf(ss), ret);
	    pos = 0;
	}
    }

    String::iter(append, s);

    ret = Cons(Leaf("\"" + String::sub(tmp, 0, pos) + "\""), ret);
    
    return List::rev(ret);
}

int len(tstring s)
{
    switch s {
    case Leaf(x):
	return String::length(x);
    case Pair(_, _, l):
	return l;
    }
}

tstring add(tstring s1, tstring s2)
{
    if (len(s1) == infty || len(s2) == infty)
	return Pair(s1, s2, infty);
    else
	return Pair(s1, s2, len(s1) + len(s2));
}

tstring concat(<tstring>list lst)
{
    switch lst {
    case Cons(hd, Nil):
	return hd;
    case Cons(hd, tl):
	return add(hd, concat(tl));
    case Nil:
	return Leaf("");
    }
}

tstring concat_at(int lev, <tstring>list lst, int limit)
{
    int sum = 0;
   
    List::iter(fun (s) {
		    if (len(s) == infty) sum = infty; 
		    else		 sum = sum + len(s);
	       }, lst);
   
    if (sum != infty)
	sum = sum + List::length(lst) - 1;
    if (sum < 0)
	return Leaf("");
    
    if (sum <= limit) {
	switch lst {
	case Cons(hd, tl):
	    _ ret = hd;
	    List::iter(fun (s) {ret = add(add(ret, Leaf(" ")), s);}, tl);
	    return ret;
	    
	case Nil:
	    return Sys::abort();
	}
    } else {
	_ promote(_ s)
	{
	    if (len(s) == infty)
		return s;
	    else
		return concat([get_indent(lev), s, Leaf("\n")]);
	}

	_ ret = concat(List::map(promote, lst));
	
	// set proper length
	return Pair(ret, Leaf(""), infty);
    }
}

tstring mkterm(int lev, string name, <tstring>list children)
{
    string end = "";
    
    if (name == "compound") {
	name = name + " [";
	end = "]";
    }
	
    switch children {
    case Cons(_, _):
	_ limit = line_len - (String::length(name) + 3 + lev);
	_ c = concat_at(lev + 1, children, limit);
	if (len(c) == infty) {
	    let opn = add(get_indent(lev), Leaf("(" + name + "\n")) in
	    let cls = add(get_indent(lev), Leaf(end + ")\n")) in
	    return Pair(opn, add(c, cls), infty);
	} else {
	    return add(Leaf("(" + name + " "), add(c, Leaf(end + ")")));
	}
	
    case Nil:
	return Leaf("(" + name + ")");
    } 
}

tstring mklist(int lev, <tstring>list args)
{
    _ limit = line_len - (lev + 2);
    _ c = concat_at(lev + 1, args, limit);

    if (len(c) == infty) {
	return concat([get_indent(lev), Leaf("[\n"), 
			    c, 
	               get_indent(lev), Leaf("]\n")]);
    } else {
	return concat([Leaf("["), c, Leaf("]")]);
    }
}

tstring convert_ty(int lev, ty t)
{
    switch t {
    case Void:		return Leaf("(void)");
    case Bool:		return Leaf("(bool)");
    case Ellipsis:	return Leaf("(ellipsis)");
    case Poly:		return Leaf("(poly)");
    case Tyref(s):	return Leaf("(tyref " + s + ")");
    case Int(s):	return Leaf("(int " + itoa(s) + ")");
    case Uint(s):	return Leaf("(uint " + itoa(s) + ")");
    case Float(s):	return Leaf("(float " + itoa(s) + ")");
    case Struct(n):	return Leaf("(struct " + n + ")");
    case Union(n):	return Leaf("(union " + n + ")");
    case Macro_type(n):	return Leaf("." + n);
    
    case Ptr(x):
	return mkterm(lev, "ptr", [convert_ty(lev + 1, x)]);
	
    case Reference(x):
	return mkterm(lev, "reference", [convert_ty(lev + 1, x)]);
	
    case Func_ptr(ret, args):
	let r = convert_ty(lev + 1, ret) in
	let a = List::map(partial(convert_ty, lev + 1), args) in
	return mkterm(lev, "func_ptr", [r, mklist(lev, a)]);
    }
}

tstring convert_attr(int lev, attr a)
{
    switch a {
    case Packed:		return Leaf("(packed)");
    case Local:			return Leaf("(local)");
    case Public:		return Leaf("(public)");
    case Extern:		return Leaf("(extern)");
    case Suppress_debug:		return Leaf("(suppress_debug)");
    case Const_func:		return Leaf("(const_func)");
    case Pure_func:		return Leaf("(pure_func)");
    case Malloc_like:		return Leaf("(malloc_like)");
    case No_return:		return Leaf("(no_return)");
    case Artificial:		return Leaf("(artificial)");
    case Weak:			return Leaf("(weak)");
    case One_only:		return Leaf("(one_only)");
    case Comdat:		return Leaf("(comdat)");
    case Aligned(i):		return Leaf("(aligned " + itoa(i) + ")");
    case Alias_set(id):		return Leaf("(alias_set " + id + ")");
    case Section(id):		return Leaf("(section " + id + ")");
    case Export_name(name):	
	return mkterm(lev + 1, "export_name", convert_string(name));
    }
}


string binop_name(binop op)
{
    switch op {
    case Eq:		return "eq";
    case Ne:		return "ne";
    case Ge:		return "ge";
    case Gt:		return "gt";
    case Le:		return "le";
    case Lt:		return "lt";
    case Plus:		return "plus";
    case Minus:		return "minus";
    case Mult:		return "mult";
    case R_div:		return "rdiv";
    case Exact_div:	return "exact_div";
    case Round_div:	return "round_div";
    case Floor_div:	return "floor_div";
    case Ceil_div:	return "ceil_div";
    case Trunc_div:	return "trunc_div";
    case Trunc_mod:	return "trunc_mod";
    case Round_mod:	return "round_mod";
    case Floor_mod:	return "floor_mod";
    case Ceil_mod:	return "ceil_mod";
    case Bit_and:	return "bit_and";
    case Bit_ior:	return "bit_ior";
    case Bit_and_tc:	return "bit_andtc";
    case Bit_xor:	return "bit_xor";
    case Max:		return "max";
    case Min:		return "min";
    case L_rotate:	return "lrotate";
    case L_shift:	return "lshift";
    case R_rotate:	return "rrotate";
    case R_shift:	return "rshift";
    case Post_inc:	return "postinc";
    case Pre_inc:	return "preinc";
    case Post_dec:	return "postdec";
    case Pre_dec:	return "predec";
    case Truth_and_if:	return "truth_andif";
    case Truth_or_if:	return "truth_orif";
    case Truth_and:	return "truth_and";
    case Truth_or:	return "truth_or";
    case Truth_xor:	return "truth_xor";
    }
}

string unop_name(unop op)
{
    switch op {
    case Negate:	return "negate";
    case Bit_not:	return "bit_not";
    case Truth_not:	return "truth_not";
    case Abs:		return "abs";
    case Addr:		return "addr";
    case R_addr:		return "raddr";
    case Ffs:		return "ffs";
    case Non_lvalue:	return "non_lvalue";
    }
}

tstring convert_expr(int lev, expr ex)
{
    tstring self(expr e)
    {
	return convert_expr(lev + 1, e);
    }
    
    tstring build_call(string name, tstring fnc, <expr>list args)
    {
	let targs = List::map(self, args) in
	return mkterm(lev, name, [fnc, mklist(lev + 1, targs)]);
    }


    tstring ty(ty t)
    {
	return convert_ty(lev + 1, t);
    }
    
    tstring ts(string name, <tstring>list args)
    {
	return mkterm(lev, name, args);
    }

    switch ex {
    case Int_const(t, v):
	return ts("const", [ty(t), Leaf(Int64::to_string(v, 10))]);
			   
    case Float_const(t, v):
	return ts("const", [ty(t), Leaf(ftoa(v))]);
			  
    case Bool_const(true):
	return Leaf(".true");
	
    case Bool_const(false):
	return Leaf(".false");
    
    case String_const(s):
	return ts(".string", convert_string(s));

    case Null(t):
	return ts("const", [ty(t), Leaf("0")]);

    case Var(t, n):
	return ts("var", [ty(t), Leaf(n)]);
    
    case Static_var(t, n):
	return ts("static_var", [ty(t), Leaf(n)]);

    case Bind(e):
	return ts("bind", [self(e)]);
    
    case Loop(e):
	return ts("loop", [self(e)]);
    
    case Ref(n):
	return Leaf("(ref " + n + ")");
	
    case Enum_ref(n):
	return Leaf("(enumref " + n + ")");
	
    case Func_ref(n):
	return Leaf("(func_ref " + n + ")");
	
    case Modify(e1, e2):
	return ts("modify", [self(e1), self(e2)]);
			   
    case Component(e, n):
	return ts("component", [self(e), Leaf(n)]);
			  
    case Arrow(e, n):
	return convert_expr(lev, Component(Indirect(e), n));

    case Return(e):
	switch e {
	case Nop:
	    return Leaf("(return (nop))");
	case Modify(Result, _):
	    return ts("return", [self(e)]);
	case _:
	    return convert_expr(lev, Return(Modify(Result, e)));
	}

    case Size_of(t):
	return ts("size_of", [ty(t)]);
	
    case Compound(lst):
	return ts("compound", List::map(self, lst));
    
    case Nop:
	return Leaf("(nop)");
    
    case Result:
	return Leaf("(result)");
    
    case Indirect(e):
	return ts("indirect", [self(e)]);
	
    case Cond(e1, e2, e3):
	return ts("cond", [self(e1), self(e2), self(e3)]);
    
    case Call(n, args):
	return build_call("call", Leaf(n), args);
	
    case F_call(n, args):
	return build_call("fcall", Leaf(n), args);
	
    case Ptr_call(e, args):
	return build_call("ptr_call", self(e), args);
	
    case Ptr_f_call(e, args):
	return build_call("ptr_fcall", self(e), args);
	
    case Built_in(n, args):
	return build_call("builtin", Leaf(n), args);
	
    case F_built_in(n, args):
	return build_call("fbuiltin", Leaf(n), args);
    
    case Convert(t, e):
	return ts("convert", [ty(t), self(e)]);

    case Block(n, e):
	return ts("block " + n, [self(e)]);

    case Exit_block(n):
	return Leaf("(exit " + n + ")");
	
    case Binary(op, e1, e2):
	return ts(binop_name(op), [self(e1), self(e2)]);
	
    case Unary(op, e):
	return ts(unop_name(op), [self(e)]);
    
    case Try_finally(e1, e2):
	return ts("try_finally", [self(e1), self(e2)]);
    
    case Line(no, e):
	return ts("wfl " + itoa(no), [self(e)]);
    
    case Comment(s, e):
	return force_newlines(lev, [Leaf("// " + s), self(e)]);
    }
}

tstring convert_decl(decl d)
{
    int cnt = 0;
    
    tstring convert_attrs(<attr>list a)
    {
	return mklist(1, List::map(partial(convert_attr, 2), a));
    }

    tstring convert_unnamed_parm(ty tt)
    {
	switch tt {
	case Ellipsis:
	    return Leaf("(ellipsis)");
	case _:
	    return mkterm(1, "var", [convert_ty(2, tt), 
				     Leaf("_" + itoa(cnt++))]);
	}
    }

    tstring convert_named_parm(*(ty, id) p)
    {
	let (t, n) = p in
	return mkterm(1, "var", [convert_ty(2, t), Leaf(n)]);
    }

    tstring convert_enum_fld(*(id, int) fld)
    {
	let (n, v) = fld in
	return Leaf("(entry " + n + " " + itoa(v) + ")");
    }

    tstring lst(<tstring>list l)
    {
	return mklist(1, l);
    }
    
    switch d {
    case Func_proto(attrs, ret_type, name, arg_types):
	return mkterm(0, "func", 
		     [convert_attrs(attrs),
		      convert_ty(1, ret_type),
		      Leaf(name),
		      lst(List::map(convert_unnamed_parm, arg_types)),
		      Leaf("(nil)")]);
	    
    case Func(attrs, ret_type, name, fargs, body):
	return mkterm(0, "func", 
		     [convert_attrs(attrs),
		      convert_ty(1, ret_type),
		      Leaf(name),
		      lst(List::map(convert_named_parm, fargs)),
		      convert_expr(1, optimize(-1, body))]);
    
    case Top_comment(c):
	return add(Leaf("// "), Leaf(c));
    
    case Struct_def_proto(n):
	return Leaf("(structdef " + n + " (nil))");
			  
    case Struct_def(attrs, n, l):
	return mkterm(0, "structdef", 
		      Cons(convert_attrs(attrs),
			   [Leaf(n), 
			    lst(List::map(convert_named_parm, l))]));
				
    case Union_def_proto(n):
	return Leaf("(uniondef " + n + " (nil))");
			  
    case Union_def(attrs, n, l):
	return mkterm(0, "uniondef", 
		      Cons(convert_attrs(attrs),
			   [Leaf(n), 
			    lst(List::map(convert_named_parm, l))]));
				
    case Ty_alias(attrs, n, t):
	return mkterm(0, "tyalias", [convert_attrs(attrs),
				     convert_named_parm((t, n))]);
			
    case Abstract_type(attrs, n):
	return mkterm(0, "abstract_type", [convert_attrs(attrs), Leaf(n)]);
			
    case Enum_def(l):
	return mkterm(0, "enumdef", [mklist(1, List::map(convert_enum_fld, l))]);
		
    case Global(attrs, t, n):
	return mkterm(0, "global", [convert_attrs(attrs),
				    convert_named_parm((t, n))]);
			
    case Input_file(s):
	return mkterm(0, "input_file", convert_string(s));
    }
}

tstring force_newlines(int lev, <tstring>list lst)
{
    _ promote(_ s)
    {
	if (len(s) == infty)
	    return s;
	else
	    return concat([get_indent(lev), s, Leaf("\n")]);
    }

    return concat(List::map(promote, lst));
}

tstring get_indent(int n)
{
    if (n >= line_len)
	n = line_len - 1;
	
    return Array::get(indents, n);
}

void print_tstring(Stdio::FILE f, tstring s)
{
    switch s {
    case Pair(s1, s2, _):
	print_tstring(f, s1);
	print_tstring(f, s2);
    case Leaf(str):
	Stdio::fputs(f, str);
    }
}

void output_decl(Stdio::FILE out, decl d)
{
    print_tstring(out, force_newlines(0, [convert_decl(d)]));
}

section init 
{
    fast_writer = false;
    infty = 0xffffff;
    line_len = 75;
    
    indents = Array::init(line_len, fun (s) (Leaf(String::make(s, ' '))));
}

// peep-hole optimizer for Ksi trees.
expr optimize(int current_line, expr ex)
{
    expr self(expr e)
    {
	return optimize(current_line, e);
    }
    
    <expr>list map_self(<expr>list l)
    {
	return List::map(self, l);
    }
    
    switch ex {
    case Int_const(_, _):
    case Float_const(_, _):
    case Bool_const(_):
    case String_const(_):
    case Null(_):
    case Var(_, _):
    case Static_var(_, _):
    case Ref(_):
    case Enum_ref(_):
    case Func_ref(_):
    case Size_of(_):
    case Nop:
    case Result:
    case Exit_block(_):
	return ex;

    case Bind(e):
	switch self(e) {
	case Bind(x):
	    return Bind(x);
	case x:
	    return Bind(x);
	}
    
    case Loop(e):
	return Loop(self(e));
    
    case Modify(e1, e2):
	return Modify(self(e1), self(e2));
			   
    case Component(e, n):
	return Component(self(e), n);
			  
    case Arrow(e, n):
	return self(Component(Indirect(e), n));

    case Return(e):
	switch e {
	case Nop:
	    return ex;
	case Modify(Result, x):
	    return Return(Modify(Result, self(x)));
	case x:
	    return self(Return(Modify(Result, self(x))));
	}
	
    case Compound(lst):
	_ ret = Nil;
	void process_one(expr e)
	{
	    switch self(e) {
	    case Compound(l):
		List::iter(fun (x) {ret = Cons(x, ret);}, l);
	    case Nop:
		skip;
	    case x:
		ret = Cons(x, ret);
	    }
	}
	
	List::iter(process_one, lst);

	switch ret {
	case Nil:
	    return Nop;
	case Cons(x, Nil):
	    return x;
	case _:
	    return Compound(List::rev(ret));
	}
    
    case Indirect(e):
	switch self(e) {
	case Unary(Addr, x):
	case Unary(R_addr, x):
	    return x;
	case x:
	    return Indirect(x);
	}

    case Cond(cond, e1, e2):
	switch self(cond) {
	case Bool_const(true):
	    return self(e1);
	case Bool_const(false):
	    return self(e2);
	case c:
	    return Cond(c, self(e1), self(e2));
	}
    
    case Call(n, args):
	return Call(n, map_self(args));
	
    case F_call(n, args):
	return F_call(n, map_self(args));
	
    case Ptr_call(e, args):
	return Ptr_call(self(e), map_self(args));
	
    case Ptr_f_call(e, args):
	return Ptr_f_call(self(e), map_self(args));
	
    case Built_in(n, args):
	return Built_in(n, map_self(args));
	
    case F_built_in(n, args):
	return F_built_in(n, map_self(args));
    
    case Convert(t, e):
	switch self(e) {
	case Convert(_, x):
	    return Convert(t, x);
	    
	case Compound([x, Int_const(_, _)]):
	case Compound([x, Null(_)]):
	    return Convert(t, x);
	
	case x:
	    return Convert(t, x);
	}

    case Block(n, e):
	return Block(n, self(e));
	
    case Binary(op, e1, e2):
	return Binary(op, self(e1), self(e2));
	
    case Unary(Addr, e):
    // R_addr not safe here
	switch self(e) {
	case Indirect(x):
	    return x;
	case x:
	    return Unary(Addr, e);
	}
	
    case Unary(op, e):
	return Unary(op, self(e));
    
    case Try_finally(e1, e2):
	switch self(e2) {
	case Nop:
	    return self(e1);
	case x:
	    return Try_finally(self(e1), x);
	}
    
    case Line(no, e):
	if (no == current_line)
	    return self(e);
	else
	    return Line(no, optimize(no, e));
    
    case Comment(s, e):
	return Comment(s, self(e));
    }
}

// vim:sts=4:sw=4
